// wuiname.c - implement TWuiName
// Copyright (c) 1993 by Ron Burk

#include <ctype.h>
#include <string.h>
#include <windows.h>
#include "wuistd.h"
#include "wuiname.h"

/***********************************************************

class TWuiName_

TWuiName_ is the class that TWuiName objects point to.  It
consists of a string (the name) and a reference count that
indicates the number of TWuiName objects that point to it. 
TWuiName_ maintains a hash table of TWuiName_ objects, so
that there is only exactly one copy of any particular
string. 

************************************************************/

class   TWuiName_
    {
public:
    TWuiName_(const char *NewName);
   ~TWuiName_();
    int         ReferenceCount;
    char       *Name() { return Name_; }
    TWuiName_  *Next;
    static TWuiName_ *Add(const char *NewName);
void Assert();
private:
    char       *Name_;
    static  int         Hash(const char *Name);
    static  TWuiName_ **HashTable;
    enum { HASHSIZE = 37 };
    };

void TWuiName_::Assert()
    {
    ASSERTDATAPTR(HashTable);
    for(int i = 0; i < HASHSIZE; ++i)
        {
        TWuiName_  *Rover = HashTable[i];
        while(Rover != NULL)
            {
            ASSERTDATAPTR(Rover);
            ASSERTSTRINGPTR(Rover->Name_);
            ASSERT(Rover->ReferenceCount > 0);
            Rover   = Rover->Next;
            }
        }
    }

TWuiName_::TWuiName_(const char *NewName)
    :   ReferenceCount(0), Next(0)
    {   MEMBERASSERT();
    if(NewName != NULL)
        {
        Name_   = StringClone(NewName);
        ASSERTSTRINGPTR(Name_);
        }
    else
        Name_   = NULL;
    }

TWuiName_::~TWuiName_()
    {   MEMBERASSERT();
    ASSERT(ReferenceCount == 0);
    int     Index = Hash(Name());
    ASSERTDATAPTR(HashTable);
    TWuiName_ *Rover = HashTable[Index];
    if(Rover == this)
        HashTable[Index]    = this->Next;
    else
        while(Rover != NULL)
            if(Rover->Next == this)
                {
                Rover->Next = this->Next;
                break;
                }
    if(Name_ != NULL)
        {
        ASSERTSTRINGPTR(Name_);
        delete[] Name_;
        }
    }

TWuiName_ **TWuiName_::HashTable;

TWuiName_ *TWuiName_::Add(const char *NewName)
    {
    ASSERTDATAPTR(NewName);
    // Create hash table, first time through
    if(HashTable == NULL)
        {
        HashTable   = new TWuiName_ *[HASHSIZE];
        ASSERTDATAPTR(HashTable);
        for(int i = 0; i < HASHSIZE; ++i)
            HashTable[i] = NULL;
        }
    // use hash to determine which linked list to search
    int         Index = Hash(NewName);
    ASSERT(Index >= 0 && Index < HASHSIZE);
    TWuiName_  *Rover = HashTable[Index];
    // search the linked list for a match
    while(Rover != NULL)
        {
        if(!strcmp(Rover->Name(), NewName))
            return Rover;
        else
            Rover   = Rover->Next;
        }

    // This string does not already exist
    Rover               = new TWuiName_(NewName);
    ASSERT(Rover != NULL);
    Rover->Next         = HashTable[Index];
    HashTable[Index]    = Rover;
    return Rover;
    }

int TWuiName_::Hash(const char *Name)
    {
    ASSERTSTRINGPTR(Name);
    unsigned Result = *Name;

    while(*Name != '\0')
        Result ^= (Result << 3) + *Name++;

    return Result % HASHSIZE;
    }


// verify that a string is legal for a WUIMAN name
int TWuiName::LegalName(const char *Name)
    {
    if(Name != NULL)
        {
        size_t Length   = strlen(Name);
        if(Length >= TWuiName::MAXNAME)
            return FALSE;
        char    Char;
        while((Char=*Name++) != '\0' && Length > 0)
            if(Char != '_' &&  !isalnum(Char))
                return FALSE;
            else
                --Length;
        }
    return TRUE;
    }

// Constructor - convert string to TWuiName.
TWuiName::TWuiName(const char *NewName)
    {   MEMBERASSERT();
    if(NewName && *NewName)
        {
        ASSERTSTRINGPTR(NewName);
        ASSERT(strlen(NewName) < MAXNAME);
        Handle  = TWuiName_::Add(NewName);
        ASSERTDATAPTR(Handle);
        ++Handle->ReferenceCount;
        }
    else
        // we map NULL or empty strings to 0
        Handle = 0;
    Assert();
    }

// Copy constructor - just increment reference count.
TWuiName::TWuiName(const TWuiName &Other)
    {   MEMBERASSERT();
    if(Other.Handle != NULL)
        {
        ASSERTDATAPTR(Other.Handle);
        Handle  = Other.Handle;
        ++Handle->ReferenceCount;
        ASSERT(Handle != 0);
Handle->Assert();
        }
    else
        Handle = 0;
    Assert();
    }

// Copy constructor - just increment reference count.
TWuiName::TWuiName(const TWuiNameBase &Other)
    {   MEMBERASSERT();
    if(Other.Handle)
        {
        Handle  = Other.Handle;
        ++Handle->ReferenceCount;
        ASSERT(Handle != 0);
Handle->Assert();
        }
    else
        Handle = 0;
    Assert();
    }

// Assignment - standard handle class concerns.
const TWuiName &TWuiName::operator=(const TWuiName &Other)
    {
    Assert();
    if(&Other != this)  // account for self-assignment
        {
        if(Handle != NULL)
            {
            ASSERTDATAPTR(Handle);
            if(--Handle->ReferenceCount <= 0)
                delete Handle;
            }
        Handle  = Other.Handle;
        if(Handle != NULL)
            {
            ASSERTDATAPTR(Handle);
            ++Handle->ReferenceCount;
            }
        }
    Assert();
    return *this;
    }

void    TWuiName::Assert() const
    {   MEMBERASSERT();
    if(Handle != NULL)
        {
        ASSERTDATAPTR(Handle);
        ASSERTSTRINGPTR(Handle->Name());
Handle->Assert();
        }
    }

// Destructor - decrement reference count, delete if zero
TWuiName::~TWuiName()
    {
    Assert();
    if(Handle != NULL)
        {
        ASSERTDATAPTR(Handle);
        if(--Handle->ReferenceCount <= 0)
            delete Handle;
        }
    }

// GetName - convert TWuiName back into original string
char *TWuiName::GetName(char *Buffer, int MaxLength) const
    {
    Assert();
    ASSERT(MaxLength >= 0);
    if(Handle) // if not NULL
        {
        ASSERTDATAPTR(Buffer);
        StringCopy(Buffer, Handle->Name(), MaxLength);
        }
    else
        *Buffer = '\0';
    Assert();
    return Buffer;
    }


